from simple_duck_ml.layers.i_layer import ILayer
from abc import ABC, abstractmethod
from typing import Dict, Optional


class ILayerWriter(ABC):
    @abstractmethod
    def __call__(self, layer: ILayer, name: Optional[str], path: str, overwrite: bool) -> Dict[str, str]: ...



from typing import Any, Dict, Optional, Self
from abc import ABC, abstractmethod
from numpy.typing import NDArray
from duckdi import Interface
import numpy as np

@Interface(label='layer')
class ILayer(ABC):
    name: str

    @abstractmethod
    def forward(self, x: NDArray[np.float64]) -> NDArray[np.float64]: ...

    @abstractmethod
    def backward(self, delta: NDArray[np.float64]) -> NDArray[np.float64]: ...
    
    @abstractmethod
    def clean_grad(self) -> None: ...

    @abstractmethod
    def update(self, learning_rate: float = 0.01, batch_size: int = 1) -> None: ...

    @classmethod
    def load(cls, path: str) -> Self: ...

    @abstractmethod
    def save(
        self, 
        name: Optional[str]=None, 
        path: str=".", 
        overwrite: bool=True,
    ) -> Dict[str, str]: ...

    @property
    @abstractmethod
    def info(self) -> Dict[str, Any]: ...



from numpy.typing import NDArray
from abc import ABC, abstractmethod
from duckdi import Interface
import numpy as np

@Interface(label='activation')
class IActivation(ABC):
    name: str

    @abstractmethod
    def __call__(self, x: NDArray[np.float64]) -> NDArray[np.float64]: ...

    @abstractmethod
    def derivative(self, x: NDArray[np.float64]) -> NDArray[np.float64]: ...



import uuid
from simple_duck_ml.serializers.create_dir import create_dir
from simple_duck_ml.serializers.toml_io import load_toml, write_toml
from simple_duck_ml.dataset_unpacker.dataset import Dataset
from typing import Dict, List, Optional, Self, Type
from simple_duck_ml.layers.i_layer import ILayer
from simple_duck_ml.loss.i_loss import ILoss
from numpy.typing import NDArray
from tqdm import trange
from duckdi import Get
import numpy as np
import os

class Model:
    def __init__(
        self,
        layers: List[ILayer],
        loss: ILoss,
        learning_rate: float
    ) -> None:
        self.layers = layers
        self.loss = loss
        self.learning_rate = learning_rate

    def forward(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        for layer in self.layers:
            x = layer.forward(x)
        return x

    def backward(self, y_pred: NDArray[np.float64], y_true: NDArray[np.float64]) -> None:
        delta = self.loss.derivative(y_pred, y_true)
        for layer in reversed(self.layers):
            delta = layer.backward(delta)

    def update(self, batch_size: int = 1) -> None:
        for layer in self.layers:
            layer.update(self.learning_rate, batch_size)

    def _one_hot(self, label: int, num_classes: int) -> NDArray[np.float64]:
        one_hot = np.zeros((num_classes, 1), dtype=np.float64)
        one_hot[label, 0] = 1.0
        return one_hot

    def fit(
        self,
        datasets: List[Dataset],
        epochs: int = 5,
        batch_size: int = 10,
        shuffle: bool = True,
        verbose: bool = True
    ) -> None:
        num_classes = len(datasets)
        samples = [(img, d.y) for d in datasets for img in d.x]
        total_samples = len(samples)

        for epoch in trange(epochs, desc="Epoch"):
            if shuffle:
                np.random.shuffle(samples)

            total_loss = 0.0

            # Mini batch loop
            for batch_start in range(0, total_samples, batch_size):
                batch = samples[batch_start:batch_start + batch_size]
                batch_loss = 0.0

                for x, label in batch:
                    y_true = self._one_hot(label, num_classes)
                    y_pred = self.forward(x)

                    loss = self.loss(y_pred, y_true)
                    batch_loss += float(loss)
                    self.backward(y_pred, y_true)

                self.update(batch_size)
                total_loss += batch_loss

            avg_loss = total_loss / total_samples
            if verbose:
                print(f"[Epoch {epoch + 1}/{epochs}] Loss: {avg_loss:.6f}")
                print("-=" * 30)

    def save(self, name: Optional[str] = None, path: str = ".", overwrite: bool = True) -> str:
        name = str(uuid.uuid4()).replace("-", "") if name is None else name
        dir = create_dir(name, path, overwrite)

        layer_paths: List[str] = []
        for i, layer in enumerate(self.layers):
            layer_name = f"layer_{i:03d}_{layer.name}"
            layer_path = layer.save(name=layer_name, path=dir, overwrite=overwrite)
            layer_paths.append(layer_path['relative'])

            print(f"[ModelSave] Saving {layer.name} -> {layer_name}")

        model_toml_path = os.path.join(dir, "model.toml")
        write_toml(
            obj={
                "model": {
                    "learning_rate": self.learning_rate,
                    "loss": self.loss.name,
                    "layers": layer_paths,
                }
            },
            path=model_toml_path,
            overwrite=overwrite,
        )

        print(f"[ModelSave] Model saved → {model_toml_path}")
        return model_toml_path

    @classmethod
    def load(cls, path: str) -> Self:
        model = os.path.join(path, "model.toml")

        def __process_keys[T](obj: Dict, key: str, expected_type: Type[T]) -> T:
            data = obj.get(key, None)
            if data is None or not isinstance(data, expected_type):
                raise KeyError(f"Error: Could Not Process \"{key}\" key!")

            return data

        model_data = __process_keys(load_toml(model), "model", Dict)

        learning_rate = __process_keys(model_data, "learning_rate", float)
        loss_name = __process_keys(model_data, "loss", str)
        layer_paths = __process_keys(model_data, "layers", List)

        layers: List[ILayer] = []
        for raw_l_path in layer_paths:
            l_path = os.path.join(path, raw_l_path)
            l_type = __process_keys(load_toml(l_path), "layer_type", str)
            layers.append(Get(ILayer, label='layer', adapter=l_type, instance=False).load(l_path))

        loss = Get(ILoss, label="loss", adapter=loss_name)
        print(f"[ModelLoad] Loaded {len(layers)} layers, loss={loss_name}, lr={learning_rate}")
        return cls(layers=layers, loss=loss, learning_rate=learning_rate)




from simple_duck_ml.activations.i_activation import IActivation
from numpy.typing import NDArray
import numpy as np

class SoftmaxActivation(IActivation):
    name = "softmax"

    def __call__(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        x_shifted = x - np.max(x, axis=0, keepdims=True)
        exp_x = np.exp(x_shifted)
        sum_exp = np.sum(exp_x, axis=0, keepdims=True)

        return exp_x / np.clip(sum_exp, 1e-12, np.inf)

    def derivative(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        """
        Retorna o gradiente (jacobiano) do softmax.
        s deve ser o vetor já ativado (saída do softmax).
        Para uso genérico, mas em prática (com cross-entropy)
        o gradiente é (y_pred - y_true).
        """
        # If is vector
        if x.ndim == 1 or (x.ndim == 2 and x.shape[1] == 1):
            x = x.reshape(-1, 1)
            jacobian = np.diagflat(x) - np.dot(x, x.T)
            return jacobian

        # If is batch
        batch_size = x.shape[1]
        grads = np.zeros_like(x)
        for i in range(batch_size):
            si = x[:, i:i+1]
            jacobian = np.diagflat(si) - np.dot(si, si.T)
            grads[:, i:i+1] = np.sum(jacobian, axis=1, keepdims=True)
        return grads




from simple_duck_ml.activations.i_activation import IActivation
from numpy.typing import NDArray
import numpy as np

class ReLuActivation(IActivation):
    name = "relu"

    def __init__(self, max_value: float = 6.0) -> None:
        self.max_value = max_value

    def __call__(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        return np.clip(np.maximum(0, x), 0, self.max_value)

    def derivative(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        return (x > 0).astype(np.float64)




from typing import Any, Dict, Iterable, List, Optional, Self, Type
from simple_duck_ml.activations.i_activation import IActivation
from simple_duck_ml.io.toml_layer_writer import TomlLayerWriter
from simple_duck_ml.layers.i_layer import ILayer
from numpy.typing import NDArray
from duckdi import Get
import numpy as np
import os

from simple_duck_ml.serializers.tensor_io import load_tensor
from simple_duck_ml.serializers.toml_io import load_toml

class ConvolutionalLayer(ILayer):
    name = "conv"

    def __init__(
        self,
        nodes_num: int,
        kernel_shape: Iterable[int],
        activation: IActivation,
        stride: int=1,
    ) -> None:
        self.nodes_num = nodes_num
        self.kernel_shape = list(kernel_shape)

        limit = np.sqrt(2.0 / np.prod(self.kernel_shape))
        self.w = np.random.randn(nodes_num, *self.kernel_shape) * limit
        self.b = np.zeros((nodes_num, 1))

        self._grad_w: NDArray[np.float64] = np.zeros_like(self.w)
        self._grad_b: NDArray[np.float64] = np.zeros_like(self.b)

        self.stride = stride
        self.activation = activation

        self._patches: Optional[NDArray[np.float64]] = None
        self._x: Optional[NDArray[np.float64]] = None
        self._output: Optional[NDArray[np.float64]] = None
    
    def __get_patches(self, x: NDArray[np.float64], stride: int=1) -> NDArray[np.float64]:
        # Expand shape to support data (only support 3D channels :/) 
        if x.ndim == 2:
            x = np.expand_dims(x, axis=-1)

        kernel_h, kernel_w, in_channels = self.kernel_shape
        input_h, input_w, input_channels = x.shape

        # Validate channel dimensions
        if in_channels != input_channels:
            raise ValueError(
                f"Input channels ({input_channels}) must match kernel channels ({in_channels})."
            )

        # Compute output (activation map) shape
        output_h = (input_h - kernel_h) // stride + 1
        output_w = (input_w - kernel_w) // stride + 1

        # Allocate array for patches
        patches = np.zeros((output_h, output_w, kernel_h, kernel_w, in_channels))

        # Extract each patch (manual sliding window)
        for i, j in np.ndindex(output_h, output_w):
            start_i = i * stride
            start_j = j * stride
            patches[i, j] = x[start_i:start_i + kernel_h, start_j:start_j + kernel_w, :]

        return patches
            
    def forward(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        if x.ndim == 2:
            x = np.expand_dims(x, axis=-1)

        patches = self.__get_patches(x, self.stride)
        self._x = x
        self._patches = patches

        output_h, output_w = patches.shape[:2]
        kernel_h, kernel_w = self.kernel_shape[:2]
        in_channels = self.kernel_shape[-1]
        
        # Flat By Patch
        total_activation_map_elements_num = output_h * output_w
        total_kernel_elements_num = kernel_h * kernel_w * in_channels
        x_flat = patches.reshape(total_activation_map_elements_num, total_kernel_elements_num)

        # Flat Weights By Node
        w_flat = self.w.reshape(self.nodes_num, -1)

        # Forward Calc
        z = np.dot(w_flat, x_flat.T) + self.b
        
        # Resize to Activation Map Shape For Each Node
        z = z.reshape(self.nodes_num, output_h, output_w)
        
        self._output = self.activation(z)
    
        # Move the channel dimension from its current index ("0") to the last position ("-1")
        return np.moveaxis(self._output, 0, -1)

    def backward(self, delta: NDArray[np.float64]) -> NDArray[np.float64]:
        if self._output is None or self._patches is None:
            raise RuntimeError("Forward should be called before Backward")

        output_h, output_w = self._patches.shape[:2]
        kernel_h, kernel_w = self.kernel_shape[:2]
        in_channels = self.kernel_shape[-1]

        # Move the channel dimension from its current index ("1") to the first position ("0")
        if delta.shape != self._output.shape:
            delta = np.moveaxis(delta, -1, 0)

        delta *= self.activation.derivative(self._output)

        # Flat By Patch
        total_activation_map_elements_num = output_h * output_w
        total_kernel_elements_num = kernel_h * kernel_w * in_channels
        x_flat = self._patches.reshape(total_activation_map_elements_num, total_kernel_elements_num)

        # Flat Weights By Node
        w_flat = self.w.reshape(self.nodes_num, -1)

        # Flat Delta by Node
        delta_flat = delta.reshape(self.nodes_num, total_activation_map_elements_num)

        # Calc and feed Weights Gradient Map 
        self._grad_w += np.dot(delta_flat, x_flat).reshape(self.w.shape)

        # Calc and feed Bias Gradient Map Ensuring The 2D Shape Of The Gradient Map
        self._grad_b += np.sum(delta, axis=(1, 2), keepdims=False).reshape(self._grad_b.shape)

        # Calc the x gradient
        grad_x_flat = np.dot(delta_flat.T, w_flat)
        
        # Re-shape the x gradient to kernel dimensions
        grad_x = np.zeros_like(self._x)
        for idx, (i, j) in enumerate(np.ndindex(output_h, output_w)):
            # np.ndindex is the same than:
                # for i in range(output_h): 
                #       for j in range(output_w):
            patch_grad = grad_x_flat[idx].reshape(kernel_h, kernel_w, in_channels)
            grad_x[i:i+kernel_h, j:j+kernel_w, :] += patch_grad

        self._patches = None
        return grad_x

    def clean_grad(self) -> None:
        self._grad_w.fill(0)
        self._grad_b.fill(0)
   
    def update(self, learning_rate: float = 0.01, batch_size: int = 1) -> None:
        self.w -= learning_rate * (self._grad_w / batch_size)
        self.b -= learning_rate * (self._grad_b / batch_size)

        self._x = None 
        self._patches = None
        
        if self._output is not None:
           self._output.fill(0)

        self.clean_grad()

    def save(
        self, 
        name: Optional[str]=None,
        path: str=".",
        overwrite: bool=True,
    ) -> Dict[str, str]:
        writer = TomlLayerWriter()
        return writer(self, name, path, overwrite)
    
    @classmethod
    def load(cls, path: str) -> Self:
        layer_info = load_toml(path, find_on_path=True)

        def __process_keys[T](obj: Dict, key: str, expected_type: Type[T]) -> T:
            data = obj.get(key, None)
            if data is None or not isinstance(data, expected_type):
                raise KeyError(f"Error: Could Not Process \"{key}\" key!")

            return data

        activation = Get(
            IActivation,
            label='activation',
            adapter=__process_keys(layer_info, "activation", str)
        )
        kernel_shape = __process_keys(layer_info, "kernel_shape", List)
        tensors_path = __process_keys(layer_info, "tensors_path", Dict)
        nodes_num = __process_keys(layer_info, "nodes_num", int)
        stride = __process_keys(layer_info, "stride", int)

        
        absolute = __process_keys(tensors_path, "absolute", str)
        relative = __process_keys(tensors_path, "relative", str)
        path = absolute if os.path.isfile(absolute) else relative 
        tensor_info = load_tensor(path, find_on_path=True)

        w = __process_keys(tensor_info, "w", np.ndarray)
        b = __process_keys(tensor_info, "b", np.ndarray)
        
        layer = cls(nodes_num, kernel_shape, activation, stride)

        layer.w = w
        layer.b = b
        return layer
    
    @property
    def info(self) -> Dict[str, Any]:
        return {
            "metadata": {
                "layer_type": self.name,
                "kernel_shape": self.kernel_shape,
                "activation": self.activation.name,
                "nodes_num": self.nodes_num,
                "stride": self.stride,
            },
            "tensors": {
                "w": self.w,
                "b": self.b,
            }
        }



from simple_duck_ml import container as _

from simple_duck_ml.dataset_unpacker.bin_dataset_unpacker import BinDatasetUnpacker
from simple_duck_ml.dataset_unpacker.dataset import Dataset

from simple_duck_ml.layers.convolutional_layer import ConvolutionalLayer
from simple_duck_ml.layers.flatten_layer import FlattenLayer
from simple_duck_ml.layers.dense_layer import DenseLayer

from simple_duck_ml.activations.softmax_activation import SoftmaxActivation
from simple_duck_ml.activations.relu_activation import ReLuActivation

from simple_duck_ml.loss.cross_entropy_loss import CrossEntropyLoss
from simple_duck_ml.loss.mse_loss import MSELoss

from simple_duck_ml.models.model import Model

__all__ = [
    "BinDatasetUnpacker",
    "Dataset",
    "ConvolutionalLayer",
    "FlattenLayer",
    "DenseLayer",
    "SoftmaxActivation",
    "ReLuActivation",
    "CrossEntropyLoss",
    "MSELoss",
    "Model",
]



from simple_duck_ml.serializers.create_dir import create_dir
from simple_duck_ml.serializers.tensor_io import write_tensor
from simple_duck_ml.serializers.toml_io import write_toml
from simple_duck_ml.io.i_layer_writer import ILayerWriter
from simple_duck_ml.layers.i_layer import ILayer
from typing import Dict, Optional
from uuid import uuid4
import os


class TomlLayerWriter(ILayerWriter):
    def __call__(self, layer: ILayer, name: Optional[str], path: str, overwrite: bool) -> Dict[str, str]:
        name = str(uuid4()).replace("-", "") if name is None else name

        dir_path = create_dir(name, path, overwrite)
        absolute_file_path = os.path.join(dir_path, name)
        to_write = {}

        tensors = layer.info.get("tensors", None)
        if tensors is not None:
            absolute_tensor_path = write_tensor(
                tensors=tensors,
                path=absolute_file_path,
                overwrite=overwrite
            )
            to_write["tensors_path"] = { 
                "absolute": absolute_tensor_path,
                "relative": os.path.join(name, f"{name}.npz")
            }

        metadata = layer.info.get("metadata", None)
        if metadata is not None:
            to_write = { **metadata, **to_write }

        absolute_path = write_toml(
            obj=to_write,
            path=absolute_file_path,
            overwrite=overwrite,
        )
        
        return { "absolute": absolute_path, "relative": os.path.join(name, f"{name}.toml") }





import os

def create_dir(name: str, path: str, overwrite: bool):
    dir = os.path.join(path, name)
    if overwrite and os.path.exists(dir) and os.path.isdir(dir):
        os.system(f"rm -rf {dir}")

    os.makedirs(dir)

    return dir





from simple_duck_ml.io.toml_layer_writer import TomlLayerWriter
from simple_duck_ml.io.i_layer_writer import ILayerWriter
from simple_duck_ml.serializers.toml_io import load_toml
from simple_duck_ml.layers.i_layer import ILayer
from typing import Any, Dict, Optional, Self
from numpy.typing import NDArray
import numpy as np


class FlattenLayer(ILayer):
    name = "flat"

    def __init__(self) -> None:
        self._input_shape: tuple[int, ...] | None = None
        self._batch_size: int | None = None

    def forward(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        """
        Receive (C, H, W, N) or (H, W, C) or (C, H, W) and flat into (features, N)
        """
        self._input_shape = x.shape
        
        # Batch (C, H, W, N)
        if x.ndim == 4:
            self._batch_size = x.shape[-1]
            flat = x.reshape(-1, self._batch_size)

        # Implicit Batch
        elif x.ndim == 3:
            self._batch_size = 1
            flat = x.reshape(-1, 1)
        
        # Already "(features, batch)"
        elif x.ndim == 2 and x.shape[1] > 1:    
            self._batch_size = x.shape[1]
            flat = x

        else:
            flat = x.reshape(-1, 1)
            self._batch_size = 1

        return flat


    def backward(self, delta: NDArray[np.float64]) -> NDArray[np.float64]:
        if self._input_shape is None:
            raise RuntimeError("Forward deve ser chamado antes do Backward")

        # Return into original Shape (C, H, W, N) or (C, H, W)
        return delta.reshape(self._input_shape)


    def clean_grad(self) -> None:
        self._input_shape = None
        self._batch_size = None


    def update(self, learning_rate: float = 0.01, batch_size: int = 1) -> None:
        pass

    def save(
        self, 
        name: Optional[str]=None,
        path: str=".",
        overwrite: bool=True,
    ) -> Dict[str, str]:
        writer = TomlLayerWriter()
        return writer(self, name, path, overwrite)
    
    @classmethod
    def load(cls, path: str) -> Self:
        info = load_toml(path, find_on_path=True)
        
        input_shape = info.get("input_shape", None)
        if input_shape is not None:
            input_shape = tuple(input_shape)

        layer = cls()
        layer._input_shape = input_shape
        layer._batch_size = info.get("batch_size", None)

        return layer

    @property
    def info(self) -> Dict[str, Any]:
        return {
            "metadata": {
                "layer_type": self.name, 
                "input_shape": self._input_shape, 
                "batch_size": self._batch_size,
            }
        }
